#!/usr/bin/env python3
import os
import re
import base64
from time import strftime, sleep, localtime
from subprocess import Popen
from dogtail.tree import root
from dogtail.utils import config
from dogtail.rawinput import keyCombo
from qecore.utility import run
from qecore.logger import QELogger
from qecore.application import Application
from qecore.flatpak import Flatpak

log = QELogger()

class TestSandbox:
    def __init__(self, component, logging=False):
        """
        :type component: str
        :param component: Name of the component that is being tested.

        :type logging: bool
        :param logging: Turn on or off logging of this submodule.
        """

        self.logging = logging

        if self.logging:
            log.info(f"__init__(self, component={component}, logging={str(self.logging)})")

        self.component = component
        self.current_scenario = None
        self.background_color = None
        self.background_image_revert = False
        self.background_image_location = None

        self.logging_start = None
        self.screenshot_run_result = None

        self.record_video = True
        self.record_video_pid = None

        self.attach_video = True
        self.attach_video_on_pass = False

        self.attach_journal = True
        self.attach_journal_on_pass = False

        self.attach_screenshot = True
        self.failed_test = False

        self.attach_faf = True
        self.attach_faf_on_pass = True

        self.workspace_return = False

        self.set_keyring = True

        self.wait_for_stable_video = True

        self.production = True

        self.applications = []
        self.default_application = None
        self.retrieve_session_data()
        self.wait_until_shell_becomes_responsive()

        self.shell = root.application("gnome-shell")

        self.preserve_old_api()


    def before_scenario(self, context, scenario):
        """
        Actions that are to be executed before every scenario.

        :type context: <behave.runner.Context>
        :param context: Pass this object from environment file.

        :type scenario: <Scenario>
        :param scenario: Pass this object from environment file.
        """

        if self.logging:
            log.info(f"before_scenario(self, context, scenario) test: {scenario.tags[-1]}")

        self.failed_test = False

        self.current_scenario = scenario.tags[0]
        self.set_log_start_time()
        self.overview_action("hide")
        self.set_typing_delay(0.2)
        self.set_debug_to_stdout_as(False)
        self.close_yelp()
        self.close_initial_setup()
        self.copy_data_folder()
        self.set_blank_screen_to_never()

        self.set_up_embedding(context)

        if self.record_video and self.production:
            self.start_recording()

        self.detect_keyring()
        self.return_to_home_workspace()


    def after_scenario(self, context, scenario):
        """
        Actions that are to be executed after every scenario.

        :type context: <behave.runner.Context>
        :param context: Pass this object from environment file.

        :type scenario: <Scenario>
        :param scenario: Pass this object from environment file.
        """

        if self.logging:
            log.info(f"after_scenario(self, context, scenario) test: {scenario.tags[0]}")

        if scenario.status == "failed":
            self.failed_test = True

        if self.attach_screenshot and self.failed_test:
            self.capture_image()

        if self.background_image_revert:
            run(" ".join((
                "gsettings",
                "set",
                "org.gnome.desktop.background",
                "picture-uri",
                self.background_image_location
            )))

        if self.record_video:
            self.stop_recording()

        self.overview_action("hide")

        for application in self.applications:
            application.kill_application()

        if self.attach_screenshot and self.failed_test:
            self.attach_screenshot_to_report(context)

        if self.attach_journal and (self.failed_test or self.attach_journal_on_pass):
            self.attach_journal_to_report(context)

        if self.attach_video and (self.failed_test or self.attach_video_on_pass):
            self.attach_video_to_report(context)

        if self.attach_faf and (self.failed_test or self.attach_faf_on_pass):
            self.attach_abrt_link_to_report(context)


    def start_recording(self):
        """
        Start recording the video.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("start_recording(self)")

        self.display_clock_seconds()
        self.set_max_video_length_to(600)


        leftover_recording_processes = run(f"pgrep -fla qecore_start_recording").strip("\n")
        if leftover_recording_processes:
            leftover_recording_process_list = leftover_recording_processes.split("\n")
            for process in leftover_recording_process_list:
                extracted_pid = process.split(" ", 1)[0]
                run(f"sudo kill -9 {extracted_pid}")

                absolute_path_to_video = os.path.expanduser("~/Videos")
                run(f"sudo rm -rf {absolute_path_to_video}/Screencast*")


        record_video_process = Popen("qecore_start_recording", shell=True)
        self.record_video_pid = record_video_process.pid


    def stop_recording(self):
        """
        Stop recording the video.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("stop_recording(self)")

        if self.record_video_pid is not None:
            run(f"sudo kill -9 {self.record_video_pid} > /dev/null")

        self.record_video_pid = None


    def get_app(self,
                name,
                a11yAppName=None,
                desktopFileExists=True,
                desktopFileName=None,
                appProcessName=None):
        """
        This function is a wrapper over :func:`get_application` to preserve the old api name.
        This is defined as an extra function because even the method parameters were renamed to
        comply with the snake_case naming style.
        """

        if self.logging:
            log.info(" ".join((
                f"get_app(self, name={name}, a11yAppName={a11yAppName},",
                f"desktopFileExists={desktopFileExists}, desktopFileName={desktopFileName},",
                f"appProcessName={appProcessName})"
            )))

        return self.get_application(name,
                                    a11y_app_name=a11yAppName,
                                    desktop_file_exists=desktopFileExists,
                                    desktop_file_name=desktopFileName,
                                    app_process_name=appProcessName)


    def get_application(self,
                        name,
                        a11y_app_name=None,
                        desktop_file_exists=True,
                        desktop_file_name=None,
                        app_process_name=None):
        """
        Return application to be used in test.

        :type name: str
        :param name: Name of the package that provides the application.

        :type a11y_app_name: str
        :param a11y_app_name: Application's name as it appears in the a11y tree.

        :type desktop_file_exists: bool
        :param desktop_file_exists: Does desktop file of the application exist?

        :type desktop_file_name: str
        :param desktop_file_name: Application's desktop file name.

        :type app_process_name: str
        :param app_process_name: Application's name as it appears in a running process.

        :return: Application class instance
        :rtype: <qecore.application.Application>

        This function is wrapped by :func:`get_app`.
        """

        if self.logging:
            log.info(" ".join((
                f"get_application(self, name={name}, a11y_app_name={a11y_app_name},",
                f"desktop_file_exists={desktop_file_exists},",
                f"desktop_file_name={desktop_file_name},",
                f"app_process_name={app_process_name})"
            )))

        new_application = Application(name, a11y_app_name, desktop_file_exists,
                                      desktop_file_name, app_process_name,
                                      self.shell, self.session_type,
                                      self.session_desktop, self.logging)

        self.applications.append(new_application)
        self.default_application = new_application \
            if self.default_application is None else self.default_application

        return new_application


    def get_flatpak(self, flatpak_id, **kwargs):
        """
        Return flatpak to be used in test.

        :type flatpak_id: str
        :param flatpak_id: Unique name of flatpak, mandatory format: org.flathub.app

        :return: Flatpak class instance
        :rtype: <qecore.flatpak.Flatpak>
        """

        if self.logging:
            log.info(f"get_flatpak(self, flatpak_id={flatpak_id}")

        flatpak = Flatpak(flatpak_id=flatpak_id, **kwargs)
        flatpak.shell = self.shell
        self.applications.append(flatpak)
        self.default_application = self.default_application or flatpak
        return flatpak


    def wait_until_shell_becomes_responsive(self):
        """
        Give some time if shell is not yet loaded fully.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`sandbox.TestSandbox.__init__`.
        """

        if self.logging:
            log.info("wait_until_shell_becomes_responsive(self)")

        if self.session_type == "x11":
            for _ in range(60):
                if not "gnome-shell" in [x.name for x in root.applications()]:
                    sleep(0.5)
                else:
                    break
        else:
            sleep(1) # in wayland there is no way to make sure the session is loaded, so just sleep


    def retrieve_session_data(self):
        """
        Get session/system data.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`__init__`.
        """

        if self.logging:
            log.info("retrieve_session_data(self)")

        self.architecture = run("uname -m").strip("\n")

        # Distributions expected for now: self.distribution = ["Red Hat Enterprise Linux", "Fedora"]
        self.distribution = run("cat /etc/os-release | grep ^NAME=")
        self.distribution = self.distribution.split("=")[-1].strip("\n").strip("\"")

        self.session_display = run("echo $DISPLAY").strip('\n')
        if not self.session_display:
            self.session_display = run("qecore_get_active_display").strip('\n')
            os.environ["DISPLAY"] = self.session_display

        try:
            self.resolution = [int(x) for x in \
                re.findall(r"\d+x\d+", run("xrandr | grep '*'"))[0].split("x")]
        except Exception as error:
            self.resolution = f"The resolution retrieval failed for: {error}"

        self.session_desktop = run("echo $XDG_SESSION_DESKTOP").strip('\n')

        self.session_type = "x11"
        if "XDG_SESSION_TYPE" in os.environ and "wayland" in os.environ["XDG_SESSION_TYPE"]:
            self.session_type = "wayland"


    def set_up_embedding(self, context):
        """
        Set up embeding to the behave html formatter.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.

        """

        def embed_data(mime_type, data, caption):
            for formatter in context._runner.formatters:
                if "html" in formatter.name:
                    formatter.embedding(mime_type=mime_type, data=data, caption=caption)
                    return

        context.embed = embed_data


    def set_log_start_time(self):
        """
        Save time.
        Will be used to retrieve logs from journal.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("set_log_start_time(self)")

        self.logging_start = strftime("%Y-%m-%d %H:%M:%S", localtime())


    def close_yelp(self):
        """
        Close yelp application that is opened after fresh system installation.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("close_yelp(self)")

        help_process_id = run("pgrep yelp").strip("\n")
        if help_process_id.isdigit():
            run(f"kill -9 {help_process_id}")


    def close_initial_setup(self):
        """
        Close initial setup window that is opened after the first login to the system.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("close_initial_setup(self)")

        run("echo yes > ~/.config/gnome-initial-setup-done")


    def set_blank_screen_to_never(self):
        """
        Set blank screen to never. For longer tests it is undesirable for screen to lock.

        .. note::

            This method is called by :func:`before_scenario`.
            There was never need to have other options,
            we do not want the system to sleep during the test.
        """

        if self.logging:
            log.info("set_blank_screen_to_never(self)")

        run("gsettings set org.gnome.desktop.session idle-delay 0")


    def set_max_video_length_to(self, number=600):
        """
        Set maximum allowed video length. With default value for 10 minutes.

        :type number: int
        :param number: Maximum video length.

        .. note::

            This method is called by :func:`before_scenario`. You can overwrite the setting.
        """

        if self.logging:
            log.info(f"set_max_video_length_to(self, number={number})")

        run(" ".join((
            f"gsettings set",
            f"org.gnome.settings-daemon.plugins.media-keys",
            f"max-screencast-length {number}"
        )))


    def display_clock_seconds(self):
        """
        Display clock seconds for better tracking test in video.

        .. note::
            
            This method is called by :func:`before_scenario`.
            There was never need to have other options,
            as we want to see the seconds ticking during the test.
        """

        if self.logging:
            log.info("display_clock_seconds(self)")

        run("gsettings set org.gnome.desktop.interface clock-show-seconds true")


    def return_to_home_workspace(self):
        """
        Return to home workspace.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("return_to_home_workspace(self)")

        if not self.workspace_return:
            return

        keyCombo("<Super><Home>")


    def set_typing_delay(self, number):
        """
        Set typing delay so slower machines will not lose characters on type.

        :type number: int
        :param number: Time in between accepted key strokes.

        .. note::

            This method is called by :func:`before_scenario`. You can overwrite the setting.

        """

        if self.logging:
            log.info(f"set_typing_delay(self, number={number})")

        config.typingDelay = number


    def set_debug_to_stdout_as(self, true_or_false=False):
        """
        Set debugging to stdout.

        :type true_or_false: bool
        :param true_or_false: Decision if debug to stdout or not.

        .. note::

            This method is called by :func:`before_scenario`. You can overwrite the setting.
        """

        if self.logging:
            log.info(f"set_debug_to_stdout_as(self, true_or_false={true_or_false})")

        config.logDebugToStdOut = true_or_false


    def copy_data_folder(self):
        """
        Copy data/ directory content to the /tmp/ directory.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("copy_data_folder(self)")

        if os.path.isdir("data/"):
            for single_file in os.listdir("data/"):
                run(f"cp data/{single_file} /tmp/{single_file}")


    def detect_keyring(self):
        """
        Detect if keyring was setup. If not, setup the keyring with empty password.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`before_scenario`.
        """

        if self.logging:
            log.info("detect_keyring(self)")

        if not self.set_keyring:
            return

        current_user = os.path.expanduser("~")
        is_keyring_set = os.path.isfile("/tmp/keyring_set")
        is_keyring_in_place = os.path.isfile(f"{current_user}/.local/share/keyrings/default")

        if not is_keyring_set or not is_keyring_in_place:
            run(f"sudo rm -rf {current_user}/.local/share/keyrings/*")

            create_keyring_process = Popen("qecore_start_keyring", shell=True)
            self.keyring_process_pid = create_keyring_process.pid

            self.shell.child("Continue").click()
            self.shell.child("Continue").click()

            run(f"touch /tmp/keyring_set")

        run("killall qecore_start_keyring")


    def capture_image(self):
        """
        Capture screenshot after failed step.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("capture_image(self)")

        if not self.production:
            return

        self.screenshot_run_result = run("gnome-screenshot -f /tmp/screenshot.png", verbose=True)


    def attach_screenshot_to_report(self, context):
        """
        Attach screenshot to the html report upon failed test.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("attach_screenshot_to_report(self, context)")

        if not self.production:
            return

        if self.screenshot_run_result[1] != 0:
            context.embed(mime_type="text/plain",
                          data=f"Screenshot capture failed: \n{self.screenshot_run_result}\n",
                          caption="Screenshot")
        else:
            data_base64 = base64.b64encode(open("/tmp/screenshot.png", "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="image/png",
                          data=data_encoded,
                          caption="Screenshot")


    def attach_image_to_report(self, context, image=None, caption="DefaultCaption"):
        """
        Attach image to the html report upon user request.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        :type image: str
        :param image: Location of the image/png file.

        :type caption: str
        :param caption: Caption that is to be displayed in test html report.

        .. note::

            Use this to attach any image to report at any time.
        """

        if self.logging:
            log.info("attach_image_to_report(self, context)")

        if not self.production:
            return

        if os.path.isfile(image):
            data_base64 = base64.b64encode(open(image, "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="image/png",
                          data=data_encoded,
                          caption=caption)


    def attach_video_to_report(self, context):
        """
        Attach video to the html report upon failed test.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("attach_video_to_report(self, context)")

        if not (self.production and self.record_video):
            return

        absolute_path_to_video = os.path.expanduser("~/Videos")
        screencast_list = [f"{absolute_path_to_video}/{file_name}" for file_name in \
            os.listdir(absolute_path_to_video) if "Screencast" in file_name]

        video_name = f"{self.component}_{self.current_scenario}"
        absolute_path_to_new_video = f"{absolute_path_to_video}/{video_name}.webm"

        if screencast_list == []:
            context.embed(mime_type="text/plain",
                          data="No video file found.",
                          caption="Video")
        else:
            if self.wait_for_stable_video:
                self.wait_for_video_encoding(screencast_list[0])

            data_base64 = base64.b64encode(open(screencast_list[0], "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="video/webm",
                          data=data_encoded,
                          caption="Video")
            run(f"mv {screencast_list[0]} {absolute_path_to_new_video}")
            run(f"sudo rm -rf {absolute_path_to_video}/Screencast*")


    def attach_journal_to_report(self, context):
        """
        Attach journal to the html report upon failed test.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("attach_journal_to_report(self, context)")

        if not self.production:
            return

        journal_run = run(" ".join((
            f"sudo journalctl",
            f"--all",
            f"--output=short-precise",
            f"--since='{self.logging_start}'"
            f"> /tmp/journalctl_short.log"
        )), verbose=True)

        if journal_run[1] != 0:
            context.embed(mime_type="text/plain",
                          data=f"Creation of journalctl file failed: \n{journal_run}\n",
                          caption="journalctl")
        else:
            context.embed(mime_type="text/plain",
                          data=open("/tmp/journalctl_short.log", "r").read(),
                          caption="journalctl")

        run("rm /tmp/journalctl_short.log")


    def attach_abrt_link_to_report(self, context):
        """
        Attach abrt link to the html report upon detected abrt FAF report.

        :type context: <behave.runner.Context>
        :param context: Passed object.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`after_scenario`.
        """

        if self.logging:
            log.info("attach_abrt_link_to_report(self, context)")

        if not self.production:
            return

        faf_reports = set()
        abrt_directories = run("sudo ls /var/spool/abrt/ | grep ccpp-").strip("\n").split("\n")

        for abrt_directory in abrt_directories:
            try:
                reason_file = f"/var/spool/abrt/{abrt_directory}/reason"
                reported_to_file = f"/var/spool/abrt/{abrt_directory}/reported_to"

                abrt_faf_reason_run = run(f"sudo cat '{reason_file}'", verbose=True)
                abrt_faf_hyperlink_run = run(f"sudo cat '{reported_to_file}'", verbose=True)

                if abrt_faf_reason_run[1] == 0 and abrt_faf_hyperlink_run[1] == 0:
                    abrt_faf_reason = abrt_faf_reason_run[0].strip("\n")
                    abrt_faf_hyperlink = abrt_faf_hyperlink_run[0].split("ABRT Server: URL=")[-1].split("\n")[0]

                    faf_reports.add((abrt_faf_hyperlink, f"Reason: {abrt_faf_reason}"))

            except Exception as error:
                print(f"Exception caught: {error}")
                continue

        if faf_reports:
            context.embed("link", faf_reports, caption="FAF reports")


    def wait_for_video_encoding(self, file_name):
        """
        Wait until the video is fully encoded.
        This is verified by video's changing size.
        Once the file is encoded the size will not change anymore.

        :type file_name: str
        :param file_name: Video location for size verification.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`attach_video_to_report`.

        This fixes some issues with the video and most of the time the video will
        get passed with all data, in the testing this took between 2-5 seconds.
        But there still are situations when the encoding is not made in the trivial amount of time
        mostly on slower machines. Currently the hard cutoff is 60 seconds after that the wait will
        terminate and the video will get passed as is to the html report.

        This time loss is no issue with few failing tests and has huge advantage of
        having an entire video with all controling elements (sometimes the video cannot be moved
        to the middle, or does not have data abouts its length). With many failing tests this might
        add significant time to the testing time. To prevent waiting for the encoded video
        and therefore not waiting at all use::

            <qecore.sandbox.TestSandbox>.wait_for_stable_video = False
        """

        if self.logging:
            log.info(f"wait_for_video_encoding(self, file_name={file_name})")

        current_size = 0
        current_stability = 0

        iteration_cutoff = 0

        while current_stability < 30:
            new_size = os.path.getsize(file_name)
            if current_size == new_size:
                current_stability += 1
            else:
                current_stability = 0

            current_size = new_size
            sleep(0.1)

            iteration_cutoff += 1
            if iteration_cutoff > 600:
                break


    def set_background(self, color=None, background_image=None, background_image_revert=False):
        """
        Change background to a single color or an image.

        :type color: str
        :param color: String black/white to set as background color.

        :type background_image: str
        :param background_image: Image location to be set as background.

        :type background_image_revert: bool
        :param background_image_revert: Upon settings this attribute to True,
            the :func:`after_scenario` will return the background to the original state,
            after the test.

        To get the wanted color you can pass strings as follows::

            color="black"
            color="white"
            color="#FFFFFF" # or any other color represented by hexadecimal
        """

        if self.logging:
            log.info("".join((
                f"set_background(self, color={color}, ",
                f"background_image={background_image})."
            )))

        self.background_image_revert = background_image_revert

        if self.background_image_revert:
            self.background_image_location = run(f"gsettings get org.gnome.desktop.background picture-uri").strip("\n")

        if background_image:
            if "file://" in background_image:
                run(f"gsettings set org.gnome.desktop.background picture-uri {background_image}")
            else:
                run(f"gsettings set org.gnome.desktop.background picture-uri file://{background_image}")
        elif color == "white":
            run("gsettings set org.gnome.desktop.background picture-uri file://")
            run("gsettings set org.gnome.desktop.background primary-color \"#FFFFFF\"")
            run("gsettings set org.gnome.desktop.background secondary-color \"#FFFFFF\"")
            run("gsettings set org.gnome.desktop.background color-shading-type \"solid\"")
        elif color == "black":
            run("gsettings set org.gnome.desktop.background picture-uri file://")
            run("gsettings set org.gnome.desktop.background primary-color \"#000000\"")
            run("gsettings set org.gnome.desktop.background secondary-color \"#000000\"")
            run("gsettings set org.gnome.desktop.background color-shading-type \"solid\"")
        elif "#" in color:
            run("gsettings set org.gnome.desktop.background picture-uri file://")
            run(f"gsettings set org.gnome.desktop.background primary-color '{color}'")
            run(f"gsettings set org.gnome.desktop.background secondary-color '{color}'")
            run("gsettings set org.gnome.desktop.background color-shading-type \"solid\"")
        else:
            log.info(" ".join((
                f"Color '{color}' is not defined.",
                f"You can define one yourself and submit merge request or"
                f"email modehnal@redhat.com for support."
            )))


    def overview_action(self, action="hide"):
        """
        Hide or show application overview.

        :type action: str
        :param action: Hide or show application overview.

        This function takes only 'hide' and 'show' value.
        """

        if self.logging:
            log.info(f"overview_action(self, action={action})")

        if action == "hide":
            run(" ".join((
                "dbus-send",
                "--session",
                "--type=method_call",
                "--dest=org.gnome.Shell",
                "/org/gnome/Shell",
                "org.gnome.Shell.Eval",
                "string:'Main.overview.hide();'"
            )))
        elif action == "show":
            run(" ".join((
                "dbus-send",
                "--session",
                "--type=method_call",
                "--dest=org.gnome.Shell",
                "/org/gnome/Shell",
                "org.gnome.Shell.FocusSearch"
            )))
        else:
            assert False, "Unknown option"


    # Preserving older api as we have tests with those function calls
    def preserve_old_api(self):
        """
        Preserving old api as the new names of functions and attributes are snake_case complient.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`__init__`.
        """

        if self.logging:
            log.info(f"preserve_old_api(self)")

        self.recordVideo = self.record_video
        self.recordVideoPid = self.record_video_pid
        self.attachVideo = self.attach_video
        self.attachVideoOnPass = self.attach_video_on_pass
        self.attachJournal = self.attach_journal
        self.attachJournalOnPass = self.attach_journal_on_pass
        self.attachScreenshot = self.attach_screenshot
        self.failedTest = self.failed_test
        self.attachFAF = self.attach_faf
        self.attachFAFOnPass = self.attach_faf_on_pass



#!/usr/bin/env python3
from time import sleep
import shlex
import configparser
from subprocess import Popen, PIPE
from dogtail.rawinput import typeText, pressKey, keyCombo
from dogtail.tree import root
from qecore.utility import run
from qecore.logger import QELogger

log = QELogger()

class Application:
    def __init__(self, component, a11y_app_name=None, desktop_file_exists=True,
                 desktop_file_name=None, app_process_name=None, shell=None,
                 session_type=None, session_desktop=None, logging=False):
        """
        .. note::

            Do **NOT** call this by yourself.
            This class is instanced by :py:meth:`sandbox.TestSandbox.get_application`.

        :type component: str
        :param component: Component name - usually the name of the application package.

        :type a11y_app_name: str
        :param a11y_app_name: Name of application as it appears in accessibility tree.

        :type desktop_file_exists: bool
        :param desktop_file_exists: States that desktop file of given application exists or not.

        :type desktop_file_name: str
        :param desktop_file_name: Name of desktop file if not same as component.

        :type app_process_name: str
        :param app_process_name: Name of application process if not same as component.

        :type logging: bool
        :param logging: Turn on logging of this submodule. Passed from sandbox.
        """

        self.logging = logging

        if self.logging:
            log.info(" ".join((
                f"__init__(self, component={component}, a11y_app_name={a11y_app_name},",
                f"desktop_file_exists={desktop_file_exists},",
                f"desktop_file_name={desktop_file_name},",
                f"app_process_name={app_process_name}, shell='shell',",
                f"session_type={session_type}, session_desktop={session_desktop},",
                f"logging={str(logging)})"
            )))

        self.shell = shell
        self.session_type = session_type
        self.session_desktop = session_desktop
        self.pid = None
        self.instance = None
        self.component = component
        self.a11y_app_name = a11y_app_name if a11y_app_name else component
        self.desktop_file_exists = desktop_file_exists
        self.exit_shortcut = "<Control_L><Q>"
        self.kill = True
        self.kill_command = None
        self.desktop_file_name = desktop_file_name if desktop_file_name else ""
        self.app_process_name = app_process_name if app_process_name else component
        self.get_desktop_file_data()

        self.preserve_old_api()


    def get_desktop_file_data(self):
        """
        Parse desktop file.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`__init__`.
        """

        if self.logging:
            log.info(f"{self.component} get_desktop_file_data(self)")

        if self.desktop_file_exists: # zenity/gnome-shell do not have desktop file
            desktop_run = run(" ".join((
                f"rpm -qlf $(which {self.component}) |",
                f"grep /usr/share/applications/ |",
                f"grep .desktop |",
                f"grep '{self.desktop_file_name}'"
            )), verbose=True)

            if desktop_run[1] != 0:
                raise Exception(f"Desktop file of application '{self.component}' was not found.")

            desktop_file = desktop_run[0].strip("\n")
            desktop_file_config = configparser.RawConfigParser()
            desktop_file_config.read(desktop_file)

            self.name = desktop_file_config.get("Desktop Entry", "name")
            self.exec = desktop_file_config.get("Desktop Entry", "exec").split(" ")[0]


    def start_via_command(self, command=None, **kwargs):
        """
        Start application via command.

        :type command: str
        :param command: Complete command that is to be used to start application.

        :type in_session: bool
        :param in_session: Start application via command in session.
        """

        in_session = None
        for key, val in kwargs.items():
            if "session" in str(key).lower():
                in_session = val

        if self.logging:
            log.info("".join((
                f"{self.component} ",
                f"start_via_command(self, command={command}, in_session={in_session})"
            )))

        if self.is_running() and self.kill:
            self.kill_application()

        if in_session:
            pressKey("Esc")
            keyCombo("<Alt><F2>")
            sleep(0.5)
            keyCombo("<Alt><F2>")
            sleep(0.5)
            enter_a_command = self.shell.findChild(lambda x: "activate" in x.actions and x.showing)
            enter_a_command.text = command if command else self.exec
            sleep(0.5)
            pressKey("Enter")
        else:
            self.process = Popen(shlex.split(command if command else self.exec),\
                 stdout=PIPE, stderr=PIPE)

        self.wait_before_app_starts(30)
        self.instance = self.get_root()


    def start_via_menu(self):
        """
        Start application via menu.

        Starting the application by opening shell overview and typing the application
        name that is taken from the application's desktop file. This does no longer work
        in a new gnome-classic as it lost the search bar.
        """

        if self.logging:
            log.info(f"{self.component} start_via_menu(self)")

        gnome_version_present = run("rpm -q gnome-shell").strip("\n")
        if ("classic" in self.session_desktop) and ("3.32" in gnome_version_present):
            assert False, f"Application cannot be started via menu in {gnome_version_present}."

        if not self.desktop_file_exists:
            raise Exception("".join((
                f"Testing target '{self.a11y_app_name}' does not have desktop file. ",
                f"Indication of user failure!"
            )))

        if self.is_running() and self.kill:
            self.kill_application()

        run(" ".join((
            "dbus-send",
            "--session",
            "--type=method_call",
            "--dest=org.gnome.Shell",
            "/org/gnome/Shell",
            "org.gnome.Shell.FocusSearch"
        )))

        sleep(1)
        typeText(self.name)
        pressKey("Enter")

        self.wait_before_app_starts(30)
        self.instance = self.get_root()


    def close_via_shortcut(self):
        """
        Close application via shortcut.

        Closing application via shortcut Ctrl+Q. 
        If for any reason the application does not have this shortcut you can define the shortcut
        to use in class attribute :py:attr:`exit_shortcut`

        .. note::

            Format of the shortcut is a string with each key character encapsuled in < >.
            Here are  a few examples: <Ctrl><Q> <Alt><F4> <Ctrl><Shift><Q>
        """

        if self.logging:
            log.info(f"{self.component} close_via_shortcut(self)")

        if not self.is_running():
            raise Exception("".join((
                f"Application '{self.a11y_app_name}' is no longer running. ",
                f"Indication of test failure!"
            )))

        keyCombo(self.exit_shortcut)

        self.wait_before_app_closes(30)
        self.instance = None


    def already_running(self):
        """
        If application is started by other means, other than methods from this class,
        this will retrieve the root of the application.
        """

        if self.logging:
            log.info(f"{self.component} already_running(self)")

        self.wait_before_app_starts(15)
        self.instance = self.get_root()


    def get_root(self):
        """
        Get accessibility root of appllication.

        :return: Return root object of application.
        :rtype: <dogtail.tree.root.application>

        This method is used upon application start to retrieve a11y object to use in the test.
        You do not need to use this by yourself, but can be useful in some situations.
        """

        if self.logging:
            log.info(f"{self.component} get_root(self) search by '{self.a11y_app_name}'")

        return root.application(self.a11y_app_name)


    def is_running(self):
        """
        Get accessibility root of appllication.

        :return: Return state of application. Running or not.
        :rtype: bool

        This method is used by various other methods in this class to ensure correct behaviour.
        You do not need to use this by yourself, but can be useful in some situations.
        """

        if self.logging:
            log.info(f"{self.component} is_running(self)")
        try:
            for _ in range(3):
                is_running = self.a11y_app_name in [x.name for x in root.applications()]
                if is_running:
                    break
            return is_running
        except Exception:
            return False


    def get_pid_list(self):
        """
        Get list of processes of running application.

        :return: Return all running processes of application, \
            seperated by new line character, not converting to list. \
            Return nothing if application is not running.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`get_all_kill_candidates`
            when stopping the application.
        """

        if self.logging:
            log.info(f"{self.component} get_pid_list(self)")

        if self.is_running():
            return run(f"pgrep -fla {self.app_process_name}").strip("\n")

        return None


    def get_all_kill_candidates(self):
        """
        Take result of :func:`get_pid_list` and return only processes of application.

        :rtype: list
        :return: Return all IDs of application processes.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`kill_application`\
            when stopping the application.

        If kill candidate happens to have ['runtest', 'behave', 'dogtail', '/usr/bin/gnome-shell']
        in its process name. Process will not be killed.
        Return empty list if no satisfactory process was found.
        This prevents testname colliding with found process so we will not kill the test itself.
        """

        if self.logging:
            log.info(f"{self.component} get_all_kill_candidates(self)")

        application_pid_string = self.get_pid_list()
        if application_pid_string:
            application_pid_list = application_pid_string.split("\n")
        else:
            return []

        final_pid_list = []
        for process in application_pid_list:
            if not (("runtest" in process) or \
                    ("behave" in process) or \
                    ("dogtail" in process) or \
                    (process == "/usr/bin/gnome-shell")):
                extracted_pid = process.split(" ", 1)[0]
                try:
                    final_pid_list.append(int(extracted_pid))
                except ValueError:
                    pass # skip non-digits
        return final_pid_list


    def kill_application(self):
        """
        Kill application.

        This method is used by :func:`start_via_command` and :func:`start_via_menu` to ensure
        the application is not running when starting the test. So we kill all application instances
        and open a new one to test on.
        You do not need to use this by yourself, but can be useful in some situations.
        """

        if self.logging:
            log.info(f"{self.component} kill(self)")

        if self.is_running() and self.kill:
            if not self.kill_command:
                for pid in self.get_all_kill_candidates():
                    run(f"sudo kill -9 {pid} > /dev/null")
            else:
                run(self.kill_command)


    # Following two could be merged, I could not think of a nice way of doing it though.
    def wait_before_app_starts(self, time_limit):
        """
        Wait before application starts.

        :type time_limit: int
        :param time_limit: Number which signifies time before the run is stopped.
        """

        if self.logging:
            log.info("".join((
                f"{self.component} ",
                f"wait_before_app_starts(self, time_limit={time_limit})"
            )))


        for _ in range(time_limit * 10):
            if not self.is_running():
                sleep(0.1)
            else:
                return

        assert False, "".join((
            f"Application '{self.a11y_app_name}' is not running. ",
            f"Indication of test failure!"
        ))

    def wait_before_app_closes(self, time_limit):
        """
        Wait before application stops.

        :type time_limit: int
        :param time_limit: Number which signifies time before the run is stopped.
        """

        if self.logging:
            log.info("".join((
                f"{self.component} ",
                f"wait_before_app_closes(self, time_limit={time_limit}"
            )))


        for _ in range(time_limit * 10):
            if self.is_running():
                sleep(0.1)
            else:
                return

        assert False, "".join((
            f"Application '{self.a11y_app_name}' is running. ",
            f"Indication of test failure!"
        ))


    def preserve_old_api(self):
        """
        Preserving old api as the new names of functions and attributes are snake_case complient.

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`application.Application.__init__`.
        """

        self.a11yAppName = self.a11y_app_name
        self.desktopFileExists = self.desktop_file_exists
        self.exitShortcut = self.exit_shortcut
        self.desktopFileName = self.desktop_file_name
        self.appProcessName = self.app_process_name
        self.getDesktopFileData = self.get_desktop_file_data
        self.getRoot = self.get_root
        self.isRunning = self.is_running
        self.getPidList = self.get_pid_list
        self.getAllKillCandidates = self.get_all_kill_candidates



#!/usr/bin/env python3
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = """Michal Odehnal <modehnal@redhat.com>"""
__version__ = "2.1"
__copyright__ = "Copyright © 2018-2020 Red Hat, Inc."
__license__ = "GPL"
__all__ = ("application",
           "common_steps",
           "flatpak",
           "get_node",
           "image_matching",
           "logger",
           "online_accounts",
           "sandbox",
           "utility")

# For compatibility reasons as utils source code was in this file.
from qecore.utility import *



#!/usr/bin/env python3
import os
from time import sleep
import yaml
from behave import step
from dogtail.rawinput import pressKey, typeText
from qecore.utility import get_application, run


#####################################################################################
# As we cannot have online accounts yaml data in, you have to provide them yourself #
#####################################################################################


class Accounts:
    """
    **With this class all accounts are added and removed.**
    """

    def __init__(self, context):
        """
        Initiate Accounts instance.

        :type context: <behave.runner.Context>
        :param context: Context object.
        """

        self.context = context
        self.account_data = None

        try:
            # if its defined in environment.
            self.application = get_application(context, "gnome-control-center")
        except Exception:
            # if its not defined in environment.
            self.application = context.sandbox.get_app(name="gnome-control-center",
                                                       desktopFileName="gnome-control-center")


    def remove_all_accounts(self):
        """
        Remove all accounts added via gnome-online-accounts.
        """

        self.context.sandbox.detect_keyring()
        self.application.start_via_command("gnome-control-center online-accounts")

        target = self.application.instance.child("Connect to your data in the cloud")
        target_panel = target.parent.findChildren(lambda x: x.roleName == "panel")[0]
        existing_accounts = [x for x in target_panel.findChildren(lambda x: x.roleName == "label")]

        for account in existing_accounts:
            account.click()
            self.application.instance.child("Remove Account").click()

        self.application.close_via_shortcut()


    def remove_account(self, account_name):
        """
        Remove single account added via gnome-online-accounts.

        :type account_name: string
        :param account_name: Name of the account.
        """

        self.context.sandbox.detect_keyring()
        self.application.start_via_command("gnome-control-center online-accounts")

        target = self.application.instance.child("Connect to your data in the cloud")
        target_panel = target.parent.findChildren(lambda x: x.roleName == "panel")[0]
        existing_accounts = {x.name: x for x in target_panel.findChildren(lambda x:\
            x.roleName == "label")}
        try:
            existing_accounts[account_name].click()
            self.application.instance.child("Remove Account").click()
        except Exception as error:
            print(f"\nAccount '{error}' was not detected or was already removed.\n")

        self.application.close_via_shortcut()


    def account_exists(self, account_name):
        """
        Found out if there is an account added via gnome-online-accounts.

        :type account_name: string
        :param account_name: Name of the account.

        :rtype: bool
        :return: Does the account exist or not.
        """

        target = self.application.instance.child("Connect to your data in the cloud")
        target_panel = target.parent.findChildren(lambda x: x.roleName == "panel")[0]
        existing_accounts = [x.name for x in target_panel.findChildren(lambda x:\
            x.roleName == "label")]

        result = False
        for account in existing_accounts:
            if account_name in account:
                result = True
        return result


    def add_account(self, account_id, login=None, password=None, server=None, yaml_file=None):
        """
        Add account to the gnome-online-accounts.

        :type account_id: string
        :param account_id: Identification fo the account.

        :type login: string
        :param login: Login name of the account user.

        :type password: string
        :param password: Password for the login.

        :type server: string
        :param server: Server.

        :type yaml_file: string
        :param yaml_file: Location of yaml file.
        """

        self.context.sandbox.detect_keyring()
        self.account_data = AccountData(account_id, login=login, password=password,
                                        server=server, yaml_file=yaml_file)
        account_name = self.account_data.name

        self.application.start_via_command("gnome-control-center online-accounts")

        target = self.application.instance.child("Connect to your data in the cloud")
        target_panel = target.parent.findChildren(lambda x: x.roleName == "panel")[1]
        accounts_to_be_added = {x.name:\
            x for x in target_panel.findChildren(lambda x: x.roleName == "label")}
        accounts_to_be_added[account_name].click()

        account_login_dialog = self.application.instance.child(f"{account_name} Account", "dialog")

        if account_name == "Google":
            add_google_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Nextcloud":
            add_nextcloud_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Facebook":
            add_facebook_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Microsoft":
            add_microsoft_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Flickr":
            add_flickr_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Pocket":
            add_pocket_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Foursquare":
            add_foursquare_account(self.context, self.account_data, account_login_dialog)
        if account_name == "Microsoft Exchange":
            add_microsoft_exchange_account(self.context, self.account_data, account_login_dialog)

        sleep(1)
        pressKey("Esc")
        assert self.account_exists(account_name), f"Account '{account_name}' was not detected.\n"


class AccountData:
    """
    **With this class we keep and get all data about accounts.**
    """

    def __init__(self, account_id, login=None, password=None,
                 server=None, yaml_data=None, yaml_file=None):
        """
        Account class.

        :type account_id: string
        :param account_id: Identification fo the account.

        :type login: string
        :param login: Login name of the account user.

        :type password: string
        :param password: Password for the login.

        :type server: string
        :param server: Server.

        :type yaml_file: string
        :param yaml_file: Location of yaml file.

        .. note::

            This class is called by :func:`online_accounts.Accounts.add_account`.
        """

        self.account_id = account_id

        self.name = account_id
        self.login = login
        self.password = password
        self.server = server

        self.yaml_data = yaml_data
        self.yaml_file = yaml_file

        if self.yaml_data:
            self.parse_account_data()

        elif self.yaml_file:
            self.get_account_data(self.yaml_file)
            self.parse_account_data()

        elif not (self.login and self.password):
            self.get_account_data()
            self.parse_account_data()


    def get_account_data(self, yaml_file="online_accounts.yaml"):
        """
        Load yaml data to the class attribute.

        :type yaml_file: string
        :param yaml_file: Location of the yaml file.
        """

        assert os.path.isfile(yaml_file),\
            "\n".join((
                "\n\nFor online accounts you need to supply the credentials in yaml format.",
                "Expecting file 'online_accounts.yaml' in your project directory.",
                "If not found you have to provide the file yourself and use step-decorator as:.",
                "'Add \"{account_name}\" account from file \"{yaml_file}\"'\n"
            ))

        with open(yaml_file, "rb") as data:
            try:
                self.yaml_data = yaml.safe_load(data)
            except yaml.YAMLError as error:
                assert False,\
                    " ".join((
                        "Parsing YAML file was not successful.",
                        f"Please check format of accounts.yaml file: {error}"
                    ))


    def parse_account_data(self):
        """
        Parse the data from the yaml data.
        Retrieves name, login, password, server.
        """

        try:
            self.name = self.yaml_data[self.account_id]["name"]
        except Exception:
            assert False,\
                f"Name for account '{self.account_id}' is expected, but was not provided."

        try:
            self.login = self.yaml_data[self.account_id]["login"]
        except Exception:
            assert False,\
                f"Login for account '{self.account_id}' is expected, but was not provided."

        try:
            self.password = self.yaml_data[self.account_id]["password"]
        except Exception:
            assert False,\
                f"Password for account '{self.account_id}' is expected, but was not provided."

        try:
            self.server = self.yaml_data[self.account_id]["server"]
        except Exception:
            if self.name == "Nextcloud" or self.name == "Microsoft Exchange":
                assert False,\
                    f"Server for account '{self.account_id}' is expected, but was not provided."


def wait_until_progress_bar_dissapears_in(account_login_dialog):
    """
    Waiting until the progress bar from the gnome-online-accounts widget disappears.

    :type account_login_dialog: Node
    :param account_login_dialog: Node in which the progress bar is being searched in.
    """

    progress_bar = account_login_dialog.findChildren(lambda x: x.roleName == "progress bar")
    while progress_bar and progress_bar[0].showing:
        sleep(0.2)


def add_google_account(context, data, account_login_dialog):
    """
    Adding a Google account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.child(roleName="entry")
    email_entry.click()
    email_entry.text = data.login
    pressKey("Enter")

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    password_entry.text = data.password
    pressKey("Enter")

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    allow_button = account_login_dialog.child("Allow", "push button")
    while not allow_button.sensitive:
        sleep(0.2)

    pressKey("End")
    allow_button.click()

    try:
        for _ in range(10):
            if account_login_dialog.child("Allow", retry=False).showing:
                sleep(1)
    except Exception:
        pass

    context.sandbox.detect_keyring()


def add_nextcloud_account(context, data, account_login_dialog):
    """
    Adding a Nextcloud account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    server_entry = account_login_dialog.findChildren(lambda x: x.roleName == "text")[1]
    server_entry.text = data.server

    username_entry = account_login_dialog.findChildren(lambda x: x.roleName == "text")[0]
    username_entry.text = data.login

    password_entry = account_login_dialog.findChildren(lambda x: x.roleName == "password text")[0]
    password_entry.text = data.password

    account_login_dialog.child("Connect").click()

    context.sandbox.detect_keyring()


def add_facebook_account(context, data, account_login_dialog):
    """
    Adding a Facebook account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.child(roleName="entry")
    email_entry.click()
    email_entry.text = data.login

    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    password_entry.text = data.password
    pressKey("Enter")

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    continue_button = account_login_dialog.findChild(lambda x:\
        "Continue as" in x.name or x.name == "OK")
    continue_button.click()

    context.sandbox.detect_keyring()


def add_microsoft_account(context, data, account_login_dialog):
    """
    Adding a Microsoft account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.child(roleName="entry")
    email_entry.click()
    email_entry.text = data.login

    next_button = account_login_dialog.child("Next")
    next_button.click()

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    password_entry.text = data.password

    sign_in = account_login_dialog.child("Sign in")
    sign_in.click()

    context.sandbox.detect_keyring()


def add_flickr_account(context, data, account_login_dialog):
    """
    Adding a Flickr account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.child(roleName="entry")
    email_entry.click()
    email_entry.text = data.login

    next_button = account_login_dialog.child("Next")
    next_button.click()

    try:
        wait_until_progress_bar_dissapears_in(account_login_dialog)
        next_button = account_login_dialog.child("Next", retry=False)
        next_button.click()
    except Exception:
        pass

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    typeText(data.password)

    sleep(5)
    pressKey("Enter")

    authorize_button = account_login_dialog.child("OK, I'LL AUTHORIZE IT")
    while not authorize_button.sensitive:
        sleep(0.2)
    pressKey("PageDown")
    authorize_button.click()

    context.sandbox.detect_keyring()


def add_pocket_account(context, data, account_login_dialog):
    """
    Adding a Pocket account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    pressKey("PageDown")
    pressKey("PageDown")
    pressKey("PageDown")

    login_with_firefox_button = account_login_dialog.child("Continue with Firefox")
    login_with_firefox_button.click()

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    try:
        assert account_login_dialog.findChild(lambda x:\
            x.roleName == "paragraph" and x.text == "desktopqe.pocket@gmail.com", retry=False)
        pressKey("PageDown")
    except Exception:
        email_entry = account_login_dialog.child("Email", "entry")
        while not email_entry.sensitive:
            sleep(0.2)
        email_entry.click()
        email_entry.text = data.login
        account_login_dialog.child("Continue").click()

    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    password_entry.text = data.password

    sign_in = account_login_dialog.child("Sign in")
    sign_in.click()

    try:
        found_node = account_login_dialog.child("Accept")
        while found_node and not found_node.sensitive:
            sleep(0.2)
            break
        sleep(1)
        found_node.click()
    except Exception:
        pass

    try:
        found_node = account_login_dialog.child("Authorize")
        while found_node and not found_node.sensitive:
            sleep(0.2)
            break
        sleep(1)
        found_node.click()
    except Exception:
        pass

    context.sandbox.detect_keyring()


def add_foursquare_account(context, data, account_login_dialog):
    """
    Adding a FourSquare account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.child(roleName="entry")
    email_entry.click()
    email_entry.text = data.login

    password_entry = account_login_dialog.child(roleName="password text")
    password_entry.click()
    password_entry.text = data.password

    pressKey("PageDown")

    log_in_and_allow_button = account_login_dialog.child("Log in and allow")
    log_in_and_allow_button.click()

    context.sandbox.detect_keyring()


def add_microsoft_exchange_account(context, data, account_login_dialog):
    """
    Adding a Microsoft Exchange account to the gnome-online-accounts.

    :type context: <behave.runner.Context>
    :param context: Context object.

    :type data: AccountData
    :param data: Data needed: account_id, login, password, server.

    :type account_login_dialog: Node
    :param account_login_dialog: gnome-online-accounts dialog for given account.
    """

    run("sudo sed -i 's/DEFAULT/LEGACY/g' /etc/crypto-policies/config")
    run("sudo update-crypto-policies")

    wait_until_progress_bar_dissapears_in(account_login_dialog)
    email_entry = account_login_dialog.findChildren(lambda x: x.roleName == "text")[2]
    email_entry.click()
    email_entry.text = data.login

    password_entry = account_login_dialog.findChildren(lambda x: x.roleName == "password text")[0]
    password_entry.click()
    password_entry.text = data.password

    custom_button = account_login_dialog.child("Custom")
    custom_button.doActionNamed("activate")

    server_entry = account_login_dialog.findChildren(lambda x: x.roleName == "text")[0]
    server_entry.click()
    server_entry.text = data.server

    account_login_dialog.child("Connect").click()

    ignore_button = account_login_dialog.findChild(lambda x: x.name == "Ignore")
    while not ignore_button.sensitive:
        sleep(0.2)
    ignore_button.click()

    context.sandbox.detect_keyring()


@step('Add "{account_name}" account')
@step('Add "{account_name}" account from file "{yaml_file}"')
@step('Add "{account_name}" account with login "{login}" and password "{password}"')
def add_account(context, account_name, login=None, password=None, yaml_file=None):
    """
    Function with behave step decorators::

        Add "{account_name}" account
        Add "{account_name}" account from file "{yaml_file}"
        Add "{account_name}" account with login "{login}" and password "{password}"
    """

    accounts = Accounts(context)
    try:
        accounts.add_account(account_name, login=login, password=password, yaml_file=yaml_file)
    except Exception:
        run(f"rm -rf {os.path.expanduser('~')}/.local/share/webkitgtk/*")
        accounts.add_account(account_name, login=login, password=password, yaml_file=yaml_file)


@step('Remove all accounts')
def remove_accounts(context):
    """
    Function with behave step decorator::

        Remove all accounts
    """

    accounts = Accounts(context)
    try:
        accounts.remove_all_accounts()
    except Exception:
        accounts.remove_all_accounts()



#!/usr/bin/env python3
import traceback
import sys
from dogtail.tree import root
from qecore.utility import QE_DEVELOPMENT

if QE_DEVELOPMENT:
    try:
        from termcolor import colored
    except ModuleNotFoundError:
        QE_DEVELOPMENT = False


class GetNode:
    def __init__(self, context, name, role_name, description,
                 m_btn, attr, a11y_root_name, retry, expect_positive):
        """
        Initiate GetNode instance.
        Workhorse for most of the :py:mod:`common_steps`.

        :type context: <behave.runner.Context>
        :param context: Context object that is passed from common steps.

        :type name: str
        :param name: Node.name

        :type role_name: str
        :param role_name: Node.roleName

        :type description: str
        :param description: Node.description

        :type m_btn: str
        :param m_btn: Mouse click after node identification.
            Accepted values are "Left", "Middle" and "Right".

        :type attr: str
        :param attr: Node identification: attribute.
            The most used options are: ["showing", "visible", "checked", "focused", "sensitive"]

        :type intance_root_name: str
        :param intance_root_name: Application name.
            Application name to be found in context.<app_name> or in accessibility tree.
            If search of accessibility tree fails the context object will be examined
            for Application instance.

        :type retry: bool
        :param retry: Option to give search function to look again a few times if search fails.
            Used for slower applications. User might want to click right away but window can
            have a few secods delay.

        :type expect_positive: bool
        :param expect_positive: Option to pass the common step call if the node is not found.
            Some steps might want the node not to be found.

        .. note::

            This class serves only for the purposes of the :py:mod:`common_steps` implementation.
        """

        a11y_roots = {x.name: x for x in root.applications()}

        if a11y_root_name in a11y_roots:
            self.root = a11y_roots[a11y_root_name]
        elif a11y_root_name is not None and hasattr(context, a11y_root_name):
            self.root = getattr(context, a11y_root_name).instance
        else:
            self.root = None

        if self.root is None:
            try:
                self.root = context.sandbox.default_application.instance
            except AttributeError:
                assert False, "".join((
                    "\nYou need to define a default application ",
                    "if you are using steps without root."
                ))

        mouse_map = {
            "Left": 1,
            "Middle": 2,
            "Right": 3,
            "None": None
        }
        try:
            self.m_btn = mouse_map[str(m_btn)]
        except KeyError:
            assert False, "\nUnknown mouse button! Check your feature file!"

        self.name = ("".join(name) if name != "Empty" else "") \
            if name != "None" else None
        self.role_name = ("".join(role_name) if role_name != "Empty" else "") \
            if role_name != "None" else None
        self.description = None if description is None else "".join(description) \
            if description != "Empty" else ""

        defined_attributes = ["showing", "visible", "checked", "focused", "sensitive"]
        self.attr = attr if attr in defined_attributes else None if attr is None else False
        assert self.attr is not False, "\n".join((
            "\nUnknown attribute. Check your feature file!",
            f"Attributes defined are {str(defined_attributes)}."
        ))

        self.retry = retry if isinstance(retry, bool) else None
        assert self.retry is not None, "\nUnknown retry state. Check your feature file!"

        self.expect_positive = expect_positive if isinstance(expect_positive, bool) else None
        assert self.expect_positive is not None, "".join((
            f"\nUnknown expect_positive state: {self.expect_positive}. ",
            f"Check your feature file!"
        ))


    def __enter__(self):
        try:
            found_node = self.root.findChild(lambda x: \
                ((not self.name is not None) or self.name in repr(x.name)) and \
                ((not self.role_name is not None) or self.role_name == x.roleName) and \
                ((not self.description is not None) or self.description in x.description) and \
                ((not self.attr is not None) or getattr(x, self.attr)) and \
                x.position[0] >= 0 and \
                x.size[0] > 0, \
                retry=self.retry)
        except Exception as error:
            if self.expect_positive:
                assert False, get_error_message(self, error)
            else:
                found_node = None
        return (self, found_node)


    def __exit__(self, exc_type, exc_value, trcb):
        if exc_type is not None:
            traceback.print_exc(file=sys.stdout)
            return False
        return True


def get_center(node):
    """
    Simple utility to get the center of the node.

    :type node: <dogtail.tree.Node>
    :param node: Node passed to the function.

    :rtype: tuple
    :return: Tuple with coordinates of the center of a Node.
    """
    return (node.position[0] + node.size[0]/2, node.position[1] + node.size[1]/2)


def get_formated_duplicates(list_size, list_of_duplicates, attr):
    """
    Take list of duplicates and format them for error message.

    :type list_size: int
    :param list_size: Size of the list_of_duplicates.

    :type list_of_duplicates: list
    :param list_of_duplicates: List of Nodes to handle for error message.

    :type attr: string
    :param attr: Node passed to the function.

    :rtype: string
    :return: Formatted string output.

    .. note::

        This serves only for the purpose of formatted error message upon search fail.
        Used by :py:func:`get_error_message`.
    """

    return "".join(sorted(set(["\t{0}: '{1}' {2}: '{3}' {4}: '{5}' {6}: '{7}' {8}: '{9}'\n".format(
        colored("name", "yellow") if QE_DEVELOPMENT else "name", repr(node.name),
        colored("roleName", "yellow") if QE_DEVELOPMENT else "roleName", node.roleName,
        colored("position", "yellow") if QE_DEVELOPMENT else "position", node.position,
        colored("size", "yellow") if QE_DEVELOPMENT else "size", node.size,
        colored(f"{attr}", "yellow") if QE_DEVELOPMENT else f"{attr}" \
            if attr else "attribute", getattr(node, attr) if attr else "None") \
    for node in list_of_duplicates]), key=str.lower)) if list_size != 0 else "\tNone\n"


def get_formated_error_message(error,
                               node_name, same_name_items,
                               node_role_name, same_role_name_items):
    """
    Take lists of duplicates with name and roleName and format them for error message.

    :type error: string
    :param error: Error - reason why the search for Node failed.

    :type node_name: string
    :param node_name: Node.name that was searched for.

    :type same_name_items: list
    :param same_name_items: List of all items with the name node_name.

    :type node_role_name: string
    :param node_role_name: Node.roleName that was searched for.

    :type same_role_name_items: list
    :param same_role_name_items: List of all items with the roleName node_role_name.

    :rtype: string
    :return: Formatted string output of all :py:func:`get_formated_duplicates`

    .. note::

        This serves only for the purpose of formatted error message upon search fail.
        Used by :py:func:`get_error_message`.
    """

    return "".join(["\n\n{0}: {1}\n\n{2}: {3}:\n{4}\n{5}: {6}:\n{7}\n".format(
        colored("Item was not found", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Item was not found", error,
        colored("Items with name", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Items with name", repr(node_name), same_name_items,
        colored("Items with roleName", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Items with roleName", node_role_name, same_role_name_items)])


def get_error_message(node, error):
    """
    Main handler for error message with :py:func:`get_formated_error_message` and
    :py:func:`get_formated_duplicates` being used to get desired result.

    :type node: GetNode
    :param node: Instanced GetNode to have all data needed about the error.

    :type error: string
    :param error: Error message as to why the search failed.

    .. note::

        This serves only for the purpose of formatted error message upon search fail
        when using :py:mod:`common_steps`.
    """

    nodes_with_name = node.root.findChildren(lambda x: \
        node.name in x.name and (not (node.name != "") or x.name != ""))
    nodes_with_name_size = len(nodes_with_name)
    nodes_with_name_formatted = get_formated_duplicates(nodes_with_name_size,
                                                        nodes_with_name,
                                                        node.attr)

    nodes_with_role_name = node.root.findChildren(lambda x: x.roleName == node.role_name)
    nodes_with_role_name_size = len(nodes_with_role_name)
    nodes_with_role_name_formatted = get_formated_duplicates(nodes_with_role_name_size,
                                                             nodes_with_role_name,
                                                             node.attr)

    return get_formated_error_message(error,
                                      node.name,
                                      nodes_with_name_formatted,
                                      node.role_name,
                                      nodes_with_role_name_formatted)



#!/usr/bin/env python3
import logging
import inspect

# Solution provided by https://stackoverflow.com/a/54209647

class Singleton(type):
    """
    Singletop class used as metaclass by :py:class:`logger.QELogger`.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class QELogger(metaclass=Singleton):
    logger = None

    def __init__(self):
        """
        Initiating logger class with some basic logging setup.
        """

        logging.basicConfig(
            level=logging.INFO,
            format="%(message)s",
            handlers=[logging.StreamHandler()])

        self.logger = logging.getLogger("desktopqe")


    @staticmethod
    def __get_call_info():
        """
        Inspecting stack to find the exact place to pass as location of the logger called.

        :rtype: tuple
        :return: Tuple with filename and line number the logger was called from.
        """

        stack = inspect.stack()

        file_name = stack[2][1].split("/")[-1]
        line_length = stack[2][2]

        return file_name, line_length


    def info(self, message, *args):
        """
        Inspecting stack to find the exact place to pass as location of the logger called.

        :rtype: tuple
        :return: Tuple with filename and line number the logger was called from.
        """
        message = "[{}:{:3}] {}".format(*self.__get_call_info(), message)
        self.logger.info(message, *args)



#!/usr/bin/env python3
from os import walk, path
from iniparse import ConfigParser

from qecore.utility import run
from qecore.application import Application

__author__ = """Martin Krajnak <mkrajnak@redhat.com>"""


class Flatpak(Application):
    def __init__(self, flatpak_id, **kwargs):
        """
        Initiate Flatpak instance, inherits Application

        :type flatpak_id: str
        :param flatpak_id: Unique flatpak identifier, mandatory format has 2 dots.
            Param is passed to Application contructor as .component

        :param kwargs: :py:class:`application.Application` parameters.
        """

        if flatpak_id.count(".") != 2:
            raise Exception(
                f"Incorrect flatpak name {flatpak_id}, e.g.: org.gnome.gedit")

        super().__init__(component=flatpak_id, **kwargs)


    def start_via_command(self, command=None, **kwargs):
        """
        Currently the safest and only option to run flatpak due to duplicated .desktop files

        :type command: str
        :param command: Command to use to start the flatpak application.

        :param kwargs: :py:class:`application.Application.start_via_command` parameters.
        """

        command = command or f"flatpak run {self.component}"
        super().start_via_command(command=command, **kwargs)


    def kill_application(self):
        """
        Killing via 'flatpak kill <flatpak_id>', sudo for @system flatpaks
        """

        if self.is_running() and self.kill:
            run(f"sudo flatpak kill {self.component}")


    def get_desktop_file_data(self):
        """
        Provide information from .desktop file, two possible locations:
            * flatpak installed as *--user*::

                ~/.local/share/flatpak/app/<flatpak_id>/<arch>/....

            * flatpak instaled as *@system* (sudo, root)::

                /var/lib/flatpak/app/<flatpak_id>/<arch>/.....

        .. note::

            Do **NOT** call this by yourself. This method is called by :func:`__init__`.
        """

        def get_desktop_file_path(flatpak_dir):
            for root, _, files in walk(flatpak_dir):
                for f in files:
                    if '.desktop' in f:
                        return path.join(root, f)
            return None

        for pth in ['~/.local/share/flatpak/app/', '/var/lib/flatpak/app/']:
            pth = path.expanduser(pth)
            if path.isdir(f'{pth}{self.component}'):
                desktop_file = get_desktop_file_path(f'{pth}{self.component}')
                break

        if not desktop_file:
            raise Exception(f"Desktop file for {self.component} not found.")

        desktop_file_config = ConfigParser()
        desktop_file_config.read(desktop_file)

        self.name = desktop_file_config.get("Desktop Entry", "name")


    def is_running(self):
        """
        Double check if running application is really a flatpak
        """
        return super().is_running() and (self.component in run('flatpak ps'))


    @property
    def start_via_menu(self):
        """
        Since flatpaks .desktop files cannot be distinguished,
        safest option to run flatpak is start_via_command

        .. note::

            This method is not available for flatpak objects.
        """

        raise NotImplementedError("Not available for flatpak objects")


    @property
    def get_pid_list(self):
        """
        Not required, killing via flatpak kill <flatpak_id>

        .. note::

            This method is not available for flatpak objects.
        """

        raise NotImplementedError("Not available for flatpak objects")


    @property
    def get_all_kill_candidates(self):
        """
        Not required, killing via flatpak kill <flatpak_id>

        .. note::

            This method is not available for flatpak objects.
        """

        raise NotImplementedError("Not available for flatpak objects")
